/***************************************************************************
 * slink2orb.c
 * A SeedLink to Antelope ORB module
 *
 * Written by Chad Trabant, ORFEUS/EC-Project MEREDIAN
 *
 * Partially based on the SeedLink interface of the modified Comserv in
 * SeisComP written by Andres Heinloo
 ***************************************************************************/

#include <signal.h>
#include "tr.h"
#include "orb.h"
#include "Pkt.h"
#include "coords.h"
#include "bury.h"
#include "arrays.h"

#include "slinkutils.h"

static char  *version      = "3.0 (2003.016)";
static short int remap     = 0;
static short int verbose   = 0;
static short int seq_valid = 0;	        /* if any data has arrived yet */

static int     orb         = -1;	/* the ORB descriptor */
static char   *statefile   = 0;	        /* state file */
static char   *parafile    = 0;	        /* parameter file */
static char   *orbaddr     = 0;  	/* the host:port of the destination ORB */
static char   *database    = 0; 	/* the database for calibration info */
static char   *selectors   = 0; 	/* default SeedLink selectors */
static char   *dumpfile    = 0; 	/* file name for data dump */
static FILE   *outfile;	                /* the dumpfile file descriptor */

static long    selecttimeout  = 500000;	/* timeout for select() call (us) */
static int     netto          = 600;	/* network timeout (s) */
static int     netdly         = 30; 	/* network reconnect delay (s) */
static int     keepalive      = 0;  	/* interval to send keepalive/heartbeat (s) */
static double  netto_time     = 0.0;	/* network timeout time stamp */
static double  netdly_time    = 0.0;	/* network re-connect delay time stamp */
static double  keepalive_time = 0.0;	/* keepalive time stamp */
static int     savestate      = 0;	/* interval to save the state file (s) */

static SLconnection slconn;      	/* connection parameters */

/* Possible query types */
static enum {
  NoQuery,
  KeepAliveQuery
}
query_mode = NoQuery;


/* Wraps Mini-SEED to make an Antelope/SEED packet (see mseed2orbpkt.c) */
extern int  mseed2orbpkt(char *, int, char *, int, char *, double *,
			 char **, int *, int *);

static void packet_proc(int blocktype, int seqnum, char *databuf,
			int datasize, int orb);
static void info_proc(char *seedrec);
static int  parameter_proc(int argcount, char **argvec);
static int  update_seqnum(char *srcname, int seqnum);
static int  burystate(void);
static void report_environ();
static void dummy_handler(int sig);
static void term_handler(int sig);
static void usage(void);


int
main(int argc, char **argv)
{
  char databuf[2080];	   /* Room for 4 SeedLink packets */
  Dbptr dbase;	           /* Database pointer */
  SLstation *curstation;
  int curseq, blocktype, bytesread;
  int recptr         = 0;  /* Pointer for data reception */
  int sendptr        = 0;  /* Pointer for data sending */
  int sl_expect_info = 0;  /* Do we expect an INFO response? */

  /* Timinig variables */
  int netto_trig, netdly_trig, keepalive_trig, savestate_trig;
  double current_time   = 0.0;
  double previous_time  = 0.0;
  double deltatime      = 0.0;
  double savestate_time = 0.0;	/* Save state file time stamp */
  
  /* For select()ing during the read loop */
  struct timeval  select_tv;
  fd_set          select_fd;
  int             select_ret;
  
  /* For the state file */
  FILE *statefileptr;	   /* File descriptor for opening 'statefile' */
  Pf   *statepf = 0;	   /* The Pf object in the state file */
  char *seqstr;	           /* To hold the ASCII string for conversion */
  char slring[12];	   /* A net/station pair (key) for a SeedLink ring */
  
  /* Keep track of the SeedLink connection state */
  enum {
    SL_DOWN,
    SL_UP,
    SL_DATA
  }
  sl_state = SL_DOWN;
  
  /* Signal handling, use POSIX calls with standardized semantics */
  struct sigaction sa;

  sa.sa_handler = dummy_handler;
  sa.sa_flags = SA_RESTART;
  sigemptyset(&sa.sa_mask);
  sigaction(SIGALRM, &sa, NULL);
  
  sa.sa_handler = term_handler;
  sigaction(SIGINT, &sa, NULL);
  sigaction(SIGQUIT, &sa, NULL);
  sigaction(SIGTERM, &sa, NULL);
  
  sa.sa_handler = SIG_IGN;
  sigaction(SIGHUP, &sa, NULL);
  sigaction(SIGPIPE, &sa, NULL);
  
  /* Process given parameters (command line and parameter file) */
  if (parameter_proc(argc, argv) < 0)
    {
      sl_log(1, 0, "Parameter processing failed.\n");
      return -1;
    }

  /* Open the state file and read it in as a Pf */
  if (statefile)
    {
      if ((statefileptr = fopen(statefile, "r")) == NULL)
	{
	  sl_log(0, 0, "state file '%s' not found\n", statefile);
	}
      else if (pfin(statefileptr, &statepf) < 0)
	{
	  sl_log(1, 0, "pfin() failed for state file\n");
	}
      if (statefileptr != NULL)
	{
	  fclose(statefileptr);
	}
    }

  /* Point to the first station */
  curstation = slconn.stations;

  /* Go through the station/net pairs and fill in the seqnum */
  while (curstation != NULL)
    {
      snprintf(slring, sizeof(slring), "%s_%s",
	       curstation->net, curstation->sta);
      curstation->seqnum = -1;
      
      if (statepf)	/* If resurrection was successful */
	{
	  if ((seqstr = pfget_string(statepf, slring)) != 0)
	    {
	      curstation->seqnum = atoi(seqstr); /* sequence was recovered */
	      sl_log(0, 2, "Recovered sequence number for %s (Dec %d)\n",
		     slring, curstation->seqnum);
	    }
	  else /* no entry found */
	    {
	      sl_log(0, 2, "No sequence number for %s\n", slring);
	    }
	}
      curstation = curstation->next;
    }

  /* Reset the current station pointer and free the state Pf */
  curstation = slconn.stations;
  pffree(statepf);

  /* Print important parameters if verbose enough */
  if (verbose >= 3)
    report_environ();

  /* ORB setup */
  orb = orbopen(orbaddr, "w&");
  if (orb < 0)
    {
      sl_log(1, 0, "%s: orbopen() error for %s\n", Program_Name, orbaddr);
      exit(1);
    }

  /* Database setup */
  if (database)
    {
      if (dbopen(database, "r+", &dbase) == dbINVALID)
	{
	  sl_log(1, 0, "dbopen(%s) failed\n", database);
	  exit(1);
	}
      finit_db(dbase);
    }
  /* Open dumpfile if requested */
  if (dumpfile)
    {
      outfile = fopen(dumpfile, "a+");
    }

  previous_time  = sl_dtime();	   /* Initialize timing base */
  netto_trig     = -1;	           /* Init net timeout trigger to reset state */
  keepalive_trig = -1;	           /* Init keepalive trigger to reset state */
  netdly_trig    = 0;	           /* Init net delay trigger to no-delay state */
  savestate_trig = 0;	           /* Init savestate file trigger to no-save state */
  savestate_time = previous_time;  /* Init savestate time */

  slconn.link    = -1;             /* Init the link descriptor */

  /* Start the primary loop  */
  while (1)
    {
      if (slconn.link == -1)
	{
	  sl_state = SL_DOWN;
	}
      
      /* Check for network timeout */
      if (sl_state == SL_DATA && netto && netto_trig > 0)
	{
	  sl_log(0, 0, "network timeout (%ds), reconnecting in %ds\n", netto, netdly);
	  slconn.link = sl_disconnect(&slconn);
	  sl_state = SL_DOWN;
	  netto_trig = -1;
	  netdly_trig = -1;
	}

      /* Check if a keepalive packet needs to be sent */
      if (sl_state == SL_DATA && keepalive && keepalive_trig > 0)
        {
          sl_log (0, 2, "sending keepalive request\n");
          query_mode = KeepAliveQuery;
          sl_request_info (&slconn, "ID", 3);
          sl_expect_info = 1;
          keepalive_trig = -1;
        }

      /* Throttle the loop while delaying */
      if (sl_state == SL_DOWN && netdly_trig > 0)
	{
	  usleep(500000);
	}

      /* Connect to remote SeedLink */
      if (sl_state == SL_DOWN && netdly_trig == 0)
	{
	  if (sl_connect (&slconn) != -1)
	    {
	      sl_state = SL_UP;
	    }
	  netto_trig = -1;
	  netdly_trig = -1;
	}

      /* Negotiate/configure the connection */
      if (sl_state == SL_UP)
	{
	  int slconfret;

	  slconfret = sl_configlink (&slconn);
	  sl_expect_info = 0;

          if (slconfret != -1)
            {
              sl_state = SL_DATA;
              recptr = sendptr = 0;     /* initialize the data buffer pointers */
            }
          else
            {
              sl_log (1, 0, "negotiation with remote SeedLink failed\n");
              slconn.link = sl_disconnect (&slconn);
              netdly_trig = -1;
            }
        }

      /* Read incoming data */
      if (sl_state == SL_DATA)
	{
	  bytesread = 0;

	  /* Poll the server for data */
	  FD_ZERO(&select_fd);
	  FD_SET(slconn.link, &select_fd);
	  select_tv.tv_sec = 0;
	  select_tv.tv_usec = selecttimeout;

	  select_ret = select((slconn.link + 1), &select_fd, NULL, NULL,
			      &select_tv);

	  if (select_ret > 0)
	    if (!FD_ISSET(slconn.link, &select_fd))
	      {
		sl_log(1, 0, "select() reported data but socket not in set!\n");
	      }
	    else
	      {
		bytesread = sl_readdata(slconn.link, (void *) &databuf[recptr],
					sizeof (databuf) - recptr, slconn.sladdr);
	      }

	  else if (select_ret < 0)
	    {
	      sl_log(1, 0, "select() error: %s\n", strerror(errno));
	    }

	  if (bytesread < 0)       /* read() failed */
	    {
	      slconn.link = sl_disconnect(&slconn);
	      netdly_trig = -1;
	    }
	  else if (bytesread > 0)  /* data is here, process it */
	    {
	      recptr += bytesread;
	      
	      while (recptr - sendptr >= PACKSIZE + SEQSIZE)
		{

		  /* Check for an INFO packet */
                  if (!strncmp (&databuf[sendptr], "SLINFO", 6))
                    {
                      if (!sl_expect_info)
                        {
                          sl_log (1, 0, "unexpected INFO packet received, skipping\n");
                        }
                      else
                        {
                          info_proc (&databuf[sendptr + SEQSIZE]);

			  /* Is a termination indicated? */
                          if ( databuf[sendptr + SEQSIZE - 1] != '*' )
                            {
                              sl_expect_info = 0;
                            }
                        }

                      sendptr += (PACKSIZE + SEQSIZE);
                      continue; /* Skip the rest */
                    }

		  /* Check the signature and sequence number */
		  if ((curseq = sl_sequence((Seedlink_head *) & databuf[sendptr])) < 0)
		    {
		      sl_log(1, 0, "received packet with invalid SeedLink header\n");
		      recptr = sendptr;	/* ignore the remaining buffer */
		      slconn.link = sl_disconnect(&slconn);
		      netdly_trig = -1;
		      break;
		    }

		  /* Check the Mini-SEED record type */
		  if ((blocktype = sl_blocktype((Seedlink_head *) & databuf[sendptr])) == BTNUM)
		    {
		      sl_log(1, 0, "invalid Mini-SEED packet detected\n");
		      recptr = sendptr;	/* ignore the remaining buffer */
		      slconn.link = sl_disconnect(&slconn);
		      netdly_trig = -1;
		      break;
		    }

		  /* Process packet */
		  packet_proc(blocktype, curseq, &databuf[sendptr + SEQSIZE],
			      PACKSIZE, orb);

		  /* Write packet to dumpfile if defined */
		  if (dumpfile)
		    {
		      if (fwrite(&databuf[sendptr + SEQSIZE],
				 PACKSIZE, 1, outfile) <= 0)
			sl_log(0, 0, "fwrite(): error writing data to %s\n", dumpfile);
		    }

		  /* Increment the send pointer */
		  sendptr += (PACKSIZE + SEQSIZE);
		}

	      memmove(databuf, &databuf[sendptr], recptr - sendptr);
	      
	      recptr -= sendptr;
	      sendptr = 0;

	      /* Reset the timeout and keepalive timers */
              netto_trig = -1;
              keepalive_trig = -1;

	      /* Catch cases where the data stream stopped */
              if (recptr == 7 && !strncmp (&databuf[sendptr], "ERROR\r\n", 7))
                {
                  sl_log (0, 0, "SeedLink reported error with the last command\n");
                  term_handler (1);
                }

              if (recptr == 3 && !strncmp (&databuf[sendptr], "END", 3))
                {
                  sl_log (0, 0, "End of selected time window\n");
                  term_handler (1);
                }
	    }
	}

      /* Save the state file every 'savestate' seconds */
      if (statefile && savestate && savestate_trig > 0)
	{
	  if ((burystate()) < 0)
	    sl_log(1, 2, "error saving intermediate state file\n");
	  else
	    sl_log(0, 2, "intermediate state file saved\n");
	  savestate_trig = -1;
      }

      /* Update timing variables when more than a 1/4 second has passed */
      current_time = sl_dtime();
      deltatime = (current_time - previous_time);
      if (deltatime >= 0.25)
	{
          previous_time = current_time;
	  
	  /* Network timeout timing logic */
	  if (netto)
	    {
	      if (netto_trig == -1)     /* reset timer */
		{
		  netto_time = current_time;
		  netto_trig = 0;
		}
	      else if (netto_trig == 0 && (current_time - netto_time) > netto)
		{
		  netto_trig = 1;
		}
	    }

          /* Keepalive/heartbeat interval timing logic */
          if (keepalive)
            {
              if (keepalive_trig == -1) /* reset timer */
                {
                  keepalive_time = current_time;
                  keepalive_trig = 0;
                }
              else if (keepalive_trig == 0 && (current_time - keepalive_time) > keepalive)
                {
                  keepalive_trig = 1;
                }
            }

	  /* Network delay timing logic */
	  if (netdly)
	    {
	      if (netdly_trig == -1)    /* reset timer */
		{
		  netdly_time = current_time;
		  netdly_trig = 1;
		}
	      else if (netdly_trig == 1 && (current_time - netdly_time) > netdly)
		{
		  netdly_trig = 0;
		}
	    }

	  /* Save state file timing logic */
	  if (savestate && statefile)
	    {
	      if (savestate_trig == -1)
		{
		  savestate_time += (double) savestate;
		  savestate_trig = 0;
		}
	      else if (savestate_trig == 0 && (current_time - savestate_time) > savestate)
		{
		  savestate_trig = 1;
		}
	    }
	}
    }  /* End of primary loop */
  
  return 0;
}				/* End of main() */


/***************************************************************************
 * packet_proc():
 * Process a received packet based on blocktype
 ***************************************************************************/
static void
packet_proc(int blocktype, int seqnum, char *databuf, int datasize, int orb)
{
  double time;
  static char *packet;
  static int bufsize = 0;
  int       mseedret = 0;
  int         nbytes = 0;
  
  char srcname[ORBSRCNAME_SIZE];
  
  /* Process waveform data and send it on */
  if (blocktype == BTDATA)
    {
      mseedret = mseed2orbpkt(databuf, datasize, database, remap, &srcname[0],
			      &time, &packet, &nbytes, &bufsize);
			      
      if (mseedret == 0)
	{
	  
	  if (orbput(orb, srcname, time, packet, nbytes))
	    sl_log(1, 0, "orbput() failed: %s(%d)\n", srcname, nbytes);
	  
	  if (verbose >= 3)	/* Report each packet */
	    {
	      char *s;
	      sl_log(0, 0, "%s %s %d\n", srcname, s = strydtime(time), nbytes);
	      free(s);
	    }
	}
      else
	{
	  sl_log(0, 0, "record not forwared, mseed2orbpkt returned: %d\n",
		 mseedret);
	}

      /* Update the station chain entry with the latest seqnum */
      if (update_seqnum(srcname, seqnum) < 0)
	sl_log(0, 0, "Unexpected data: %s\n", srcname);

    }
}				/* End of packet_proc() */


/***************************************************************************
 * info_proc():
 * Process XML-based INFO packets.
 ***************************************************************************/
static void
info_proc(char *seedrec)
{
  Seed_record_header *seedhead = (Seed_record_header *) seedrec;

  /* Check for an error condition */
  if (!strcmp(seedhead->channel_id, "ERR"))
    {
      sl_log(1, 0, "Keepalives are not enabled on server!\n");
      return;
    }

  /* For keepalive messages (INFO ID) */
  if (query_mode == KeepAliveQuery)
    {
      sl_log(0, 2, "keepalive packet received\n");
      query_mode = NoQuery;
    }
  else
    {
      sl_log(1, 0, "Unexpected INFO packet received\n");
      query_mode = NoQuery;
    }
}				/* End of info_proc() */


/***************************************************************************
 * parameter_proc():
 * Process the command line and parameter file parameters.
 *
 * Returns 0 on success, and -1 on failure
 ***************************************************************************/
static int
parameter_proc(int argcount, char **argvec)
{
  int   stacount;
  char *tptr;
  char *key;
  char *netk, *stak;
  Pf   *pf;
  Arr  *stations = 0;
  Tbl  *stakeys;
  SLstation *curstation;
  SLstation *tmpstation;

  /* Init connection parameters */
  slconn.begin_time     = NULL;
  slconn.end_time       = NULL;
  slconn.multistation   = 0;
  slconn.server_version = 0.0;

  Program_Name = argvec[0];
  elog_init(argcount, argvec);
  elog_notify(0, "%s version %s\n", Program_Name, version);
  
  if (argcount <= 2)
    usage();

  /* Process all but last 2 command line arguments */
  for (optind = 1; optind < (argcount - 2); optind++)
    {
      if (strncmp(argvec[optind], "-v", 2) == 0)
	{
	  verbose += strspn(&argvec[optind][1], "v");
	}
      else if (strcmp(argvec[optind], "-S") == 0)
	{
	  statefile = argvec[++optind];
	}
      else if (strcmp(argvec[optind], "-pf") == 0)
	{
	  parafile = argvec[++optind];
	}
      else if (strcmp(argvec[optind], "-nt") == 0)
	{
	  netto = atoi(argvec[++optind]);
	  /*
	   * Set the time to -1 means no override by the
	   * parameter file
	   */
	  netto_time = -1.0;
	}
      else if (strcmp(argvec[optind], "-nd") == 0)
	{
	  netdly = atoi(argvec[++optind]);
	  /*
	   * Set the time to -1 means no override by the
	   * parameter file
	   */
	  netdly_time = -1.0;
	}
      else if (strcmp(argvec[optind], "-k") == 0)
	{
	  keepalive = atoi(argvec[++optind]);
	  /*
	   * Set the time to -1 means no override by the
	   * parameter file
	   */
	  keepalive_time = -1.0;
	}
      else if (strcmp(argvec[optind], "-o") == 0)
	{
	  dumpfile = argvec[++optind];
	}
      else if (strcmp(argvec[optind], "-d") == 0)
	{
	  database = argvec[++optind];
	}
      else if (strcmp(argvec[optind], "-r") == 0)
	{
	  remap = 1;
	}
      else
	{
	  elog_complain(0, "Unknown argument: %s\n", argvec[optind]);
	  usage();
	}
    }

  /* Initialize the verbosity for the sl_log function */
  sl_log(-1, verbose);

  /* For the last two required arguments */
  if ((argcount - optind) < 2)
    usage();

  for (; optind < argcount; optind++)
    {
      if (optind < (argcount - 2))
	usage();
      if (optind == (argcount - 2))
	slconn.sladdr = argvec[optind];
      if (optind == (argcount - 1))
	orbaddr = argvec[optind];
    }

  /* Check SeedLink address, if no host add 'localhost' */
  if (*slconn.sladdr == ':')
    {
      tptr = (char *) malloc(strlen(slconn.sladdr) + 10);
      sprintf(tptr, "localhost%s", slconn.sladdr);
      slconn.sladdr = tptr;
    }

  /* Read parameter file if supplied */
  if (parafile)
    {
      if ((pfread(parafile, &pf)) < 0)
	{
	  sl_log(1, 0, "error reading parameter file\n");
	  exit(1);
	}
      else
	{
	  /* Do some extra work so parameters are optional */
	  
	  if ((tptr = pfget_string(pf, "selecttimeout")) != 0)
	    selecttimeout = atol((char *) tptr);
	  
	  /* Only read network timeout if not set on command line */
	  if (netto_time != -1.0)
	    if ((tptr = pfget_string(pf, "nettimeout")) != 0)
	      netto = atoi((char *) tptr);
	  
	  /* Only read network reconnect delay if not set on command line */
	  if (netdly_time != -1.0)
	    if ((tptr = pfget_string(pf, "netdelay")) != 0)
	      netdly = atoi((char *) tptr);

	  if ((tptr = pfget_string(pf, "savestate")) != 0)
	    savestate = atoi((char *) tptr);
	  
	  if ((tptr = pfget_string(pf, "keepalive")) != 0)
	    keepalive = atoi((char *) tptr);
	  
	  if ((tptr = pfget_string(pf, "selectors")) != 0 &&
	      strlen(tptr) > 0)
	    selectors = strdup((char *) tptr);
	  
	  /* 'stations' == 0 if not found */
	  stations = pfget_arr(pf, "stations");
	}
    }

  /* Translate the 'stations' Arr, if given */
  if (stations != 0)
    {
      slconn.multistation = 1;
      stacount = 0;
      stakeys = keysarr(stations);
      if (maxtbl(stakeys) <= 0)
	{
	  sl_log(1, 0, "'stations' array in pf is empty!\n");
	  return -1;
	}
      /* Go through each entry in the 'stations' Arr */
      while ((key = (char *) shifttbl(stakeys)) != NULL)
	{
	  tmpstation = (SLstation *) malloc(sizeof(SLstation));
	  netk = strdup(key);

	  /* Parse out the net/sta pair */
	  if ((stak = strchr(netk, '_')) == NULL)
	    {
	      sl_log(1, 0, "Not in 'NET_STA' format: %s\n", netk);
	      return -1;
	    }
	  *stak++ = '\0';
	  
	  /* Check the net/sta pair */
	  if (strlen(netk) > 5 || strlen(netk) <= 0)
	    {
	      sl_log(1, 0, "Problem with network code: %s\n", key);
	      return -1;
	    }
	  if (strlen(stak) > 5 || strlen(stak) <= 0)
	    {
	      sl_log(1, 0, "Problem with station code: %s\n", key);
	      return -1;
	    }
	  /* Alloc some space and copy them into the struct */
	  tmpstation->net = (char *) malloc(6);
	  tmpstation->sta = (char *) malloc(6);
	  strncpy(tmpstation->net, netk, 6);
	  strncpy(tmpstation->sta, stak, 6);
	  
	  free(netk);

	  /* Retreive selectors for this station, if none use default */
	  tptr = (char *) getarr(stations, key);
	  if (strlen(tptr) == 0)
	    {
	      tmpstation->selectors = selectors;
	    }
	  else
	    {
	      tmpstation->selectors = strdup((char *) tptr);
	    }
	  
	  /* Add this to the list */
	  if (stacount++ == 0)
	    {
	      curstation = tmpstation;
	      slconn.stations = curstation;
	    }
	  else
	    {
	      curstation->next = tmpstation;
	      curstation = curstation->next;
	    }
	}

      /* Free the Arr and Tbl */
      freearr(stations, 0);
      freetbl(stakeys, 0);
    }
  else
    {		/* No 'stations' array, assuming uni-station mode */
      curstation = (SLstation *) malloc(sizeof(SLstation));
      curstation->selectors = selectors;
      curstation->net = strdup("XX");
      curstation->sta = strdup("UNI");
      slconn.stations = curstation;
    }

  /* Indicate the final station structure with a NULL */
  curstation->next = NULL;

  /* Free the parameter file structure */
  if (parafile)
    {
      pffree(pf);
    }
  return 0;
}				/* End of parameter_proc() */


/***************************************************************************
 * update_seqnum():
 * Search through the station chain and update the sequence number
 * for the correct entry.  'srcname' is expected to be of the form
 * 'IU_KONO_BHZ_00' such that the first two fields are the same as
 * the SeedLink ring name (as indicated in the 'stations' array in
 * the parameter file).
 *
 * Returns 0 on success and -1 if no match was found
 ***************************************************************************/
static int
update_seqnum(char *srcname, int seqnum)
{
  int        retcode = -1;
  char       srcring[12];
  char       slring[12];
  Srcname    parts;
  SLstation *curstation;
  
  curstation = slconn.stations;

  /* For uni-station mode */
  if (strcmp(curstation->net, "XX") == 0 &&
      strcmp(curstation->sta, "UNI") == 0)
    {
      curstation->seqnum = seqnum;
      retcode = 0;
    }

  /* For multi-station mode */
  else
    {
      split_srcname(srcname, &parts);
      sprintf(srcring, "%s_%s", parts.src_net, parts.src_sta);
      
      while (curstation != NULL)
	{
	  sprintf(slring, "%s_%s", curstation->net, curstation->sta);
	  if (strcmp(slring, srcring) == 0)
	    {
	      curstation->seqnum = seqnum;
	      retcode = 0;
	      break;
	    }
	  curstation = curstation->next;
	}
    }
  
  return retcode;
}				/* End of update_seqnum() */


/***************************************************************************
 * burystate():
 * Traverse the station chain, build a state file (Pf object) and write the
 * state file.
 *
 * Returns 0 on success and -1 if pfwrite() failed
 ***************************************************************************/
static int
burystate(void)
{
  int        retval = 0;
  char       seqstr[30];
  char       slring[12];
  SLstation *curstation;
  Pf        *statepf;
  
  statepf = pfnew(PFFILE);
  
  /* Traverse station chain and build Pf entries */
  curstation = slconn.stations;
  while (curstation != NULL)
    {
      sprintf(slring, "%s_%s", curstation->net, curstation->sta);
      sprintf(seqstr, "%d", curstation->seqnum);
      
      pfput_string(statepf, slring, seqstr);
      
      curstation = curstation->next;
    }

  /* Save the resulting Pf object */
  if (pfwrite(statefile, statepf) != 0)
    {
      retval = -1;
    }
  pffree(statepf);
  
  return retval;
}				/* End of burystate() */


/***************************************************************************
 * report_environ():
 * Report the state of global variables, for testing.
 ***************************************************************************/
static void
report_environ()
{
  int             stacount = 0;
  SLstation      *curstation;
  
  if (statefile)
    sl_log(0, 0, "statefile:\t%s\n", statefile);
  else
    sl_log(0, 0, "'statefile' not defined\n");
  
  sl_log(0, 0, "savestate:\t%d\n", savestate);
  
  if (parafile)
    sl_log(0, 0, "parafile:\t%s\n", parafile);
  else
    sl_log(0, 0, "'parafile' not defined\n");
  
  if (database)
    sl_log(0, 0, "database:\t%s\n", database);
  else
    sl_log(0, 0, "'database' not defined\n");
  
  sl_log(0, 0, "verbose:\t%d\n", verbose);
  sl_log(0, 0, "remap:\t%d\n", remap);
  sl_log(0, 0, "selecttimeout:\t%ld\n", selecttimeout);
  sl_log(0, 0, "nettimeout:\t%d\n", netto);
  sl_log(0, 0, "netdelay:\t%d\n", netdly);
  
  if (selectors)
    sl_log(0, 0, "selectors:\t%s\n", selectors);
  else
    sl_log(0, 0, "'selectors' not defined\n");
  
  sl_log(0, 0, "keepalive:\t%d\n", keepalive);
  
  if (orbaddr)
    sl_log(0, 0, "orbaddr:\t%s\n", orbaddr);
  else
    sl_log(0, 0, "'orbaddr' not defined\n");
  
  if (slconn.sladdr)
    sl_log(0, 0, "sladdr:\t%s\n", slconn.sladdr);
  else
    sl_log(0, 0, "'slconn.sladdr' not defined\n");
  sl_log(0, 0, "link:\t%d\n", slconn.link);
  
  sl_log(0, 0, "slconn.multistation:\t%d\n", slconn.multistation);
  
  curstation = slconn.stations;
  
  sl_log(0, 0, "'stations' array:\n");
  while (curstation != NULL)
    {
      if (curstation->net)
	sl_log(0, 0, "  %d - net: %s\n",
	       stacount, curstation->net);
      else
	sl_log(0, 0, "'net' not defined\n");
      
      if (curstation->sta)
	sl_log(0, 0, "  %d - sta: %s\n",
	       stacount, curstation->sta);
      else
	sl_log(0, 0, "'sta' not defined\n");
      
      if (curstation->selectors)
	sl_log(0, 0, "  %d - selectors: %s\n",
	       stacount, curstation->selectors);
      else
	sl_log(0, 0, "  %d - 'selectors' not defined\n", stacount);
      
      sl_log(0, 0, "  %d - seqnum: %d\n", stacount, curstation->seqnum);
      
      stacount++;
      curstation = curstation->next;
    }
}				/* End of report_environ() */


/***************************************************************************
 * term_handler() and dummy_handler():
 * Signal handler routines.  term_handler() does all appropriate clean-up
 * before exiting
 ***************************************************************************/
static void
term_handler(int sig)
{
  /* The show is over, wrap everything up */
  
  if (slconn.link != -1)
    sl_disconnect(&slconn);
  
  if (dumpfile)
    fclose(outfile);
  
  if (orb != -1)
    orbclose(orb);
  
  /* Bury the sequence number if statefile was supplied */
  if (statefile && seq_valid)
    {
      if ((burystate()) < 0)
	sl_log(1, 0, "error saving state file\n");
      else
	sl_log(0, 0, "state file saved\n");
    }
  /* flush the error log just in case */
  elog_print(stdout, 0);
  
  exit(0);
}

static void
dummy_handler(int sig)
{
}


/***************************************************************************
 * usage():
 * Print the usage message and exit.
 ***************************************************************************/
static void
usage(void)
{
  printf("\n"
	 "Usage: slink2orb [-d database] [-nd delay] [-nt timeout]\n"
	 "                 [-k interval] [-pf parameterfile] [-S statefile]\n"
	 "                 [-o dumpfile] [-r] [-v] SeedLink ORB\n"
	 "\n"
	 "Antelope Contributed Software\n"
	 "\n"
	 "Chad Trabant\n"
	 "ORFEUS/EC-Project MEREDIAN\n"
	 "\n"
	 "Please report problems to trabant@knmi.nl\n"
	 "\n");
         
  exit(1);
}
