/***************************************************************************
 * slinkutils.c
 * Routines for communicating with a SeedLink server
 *
 * Written by Chad Trabant, ORFEUS/EC-Project MEREDIAN
 *
 * Originally based on the SeedLink interface of the modified Comserv in
 * SeisComP written by Andres Heinloo
 *
 * Version: 2003.014 for slink2orb
 ***************************************************************************/

#include "slinkutils.h"
#include "elog.h"

/* Functions only used in this source file */
int sl_sayhello (SLconnection * slconn);
int sl_checkversion (SLconnection * slconn, float version);
int sl_negotiate_uni (SLconnection * slconn);
int sl_negotiate_multi (SLconnection * slconn);


/***************************************************************************
 * sl_configlink():
 * Configure/negotiate data stream(s) with the remote SeedLink
 * server.  Negotiation will be either uni or multi-station
 * depending on the value of 'multistation' in the SLconnection
 * struct.
 *
 * Returns -1 on errors, otherwise returns the link descriptor.
 ***************************************************************************/
int
sl_configlink (SLconnection * slconn)
{
  int ret = -1;

  if (slconn->multistation)
    {
      if (sl_checkversion (slconn, 2.5) >= 0)
	{
	  ret = sl_negotiate_multi (slconn);
	}
      else
	{
	  sl_log (1, 0,
		  "[%s] detected  SeedLink version (%.3f) does not support multi-station protocol\n",
		  slconn->sladdr, slconn->server_version);
	  ret = -1;
	}
    }
  else
    ret = sl_negotiate_uni (slconn);

  return ret;
}


/***************************************************************************
 * sl_sayhello():
 * Send the HELLO command and attempt to parse the server version
 * number from the returned string.  The server version is set to 0.0
 * if it can not be parsed from the returned string.
 *
 * Returns -1 on errors, 0 on success.
 ***************************************************************************/
int
sl_sayhello (SLconnection * slconn)
{
  int ret = 0;
  int pass = 0;
  int count = 0;
  int bytesread = 0;
  char sendstr[100];		/* A buffer for command strings */
  char servstr[40];		/* The remote server ident */
  char servid[10];		/* Server ID string, i.e. 'SeedLink' */
  char readbuf[100];		/* A buffer for responses */

  /* Send HELLO */
  sprintf (sendstr, "HELLO\r");
  sl_log (0, 2, "[%s] sending: HELLO\n", slconn->sladdr);
  bytesread = sl_writedata (slconn->link, (void *) sendstr,
			    strlen (sendstr) + 1, slconn->sladdr,
			    readbuf, sizeof (readbuf));
  if (bytesread < 0)
    {
      return -1;
    }

  /* Parse the server ID and version from the returned string */
  ret = sscanf (readbuf, "%s v%f ", &servid[0], &slconn->server_version);

  /* Check the response to HELLO */
  if (ret == 2 && strncasecmp (servid, "SEEDLINK", 8) == 0)
    {
      pass = 1;
    }
  else if (strncasecmp (readbuf, "SEEDLINK", 8) == 0)
    {
      pass = 1;
      slconn->server_version = 0.0;
    }

  if (pass)
    {
      count = strcspn (readbuf, "\r");
      if (count > 40)
	{
	  count = 40;
	}
      strncpy (servstr, readbuf, count);
      servstr[count] = '\0';
      sl_log (0, 1, "[%s] connected to: %s\n", slconn->sladdr, servstr);
    }
  else
    {
      sl_log (1, 0, "[%s] incorrect response to HELLO: %.*s\n",
	      slconn->sladdr, bytesread, readbuf);
      return -1;
    }

  return 0;
}				/* End of sl_sayhello() */


/***************************************************************************
 * sl_checkversion():
 * Check server version number against some value
 *
 * Returns:
 *  1 = version is greater than or equal to value specified
 *  0 = no server version is known
 * -1 = version is less than value specified
 ***************************************************************************/
int
sl_checkversion (SLconnection * slconn, float version)
{
  if (slconn->server_version == 0.0)
    {
      return 0;
    }
  else if (slconn->server_version >= version)
    {
      return 1;
    }
  else
    {
      return -1;
    }
}				/* End of sl_checkversion */


/***************************************************************************
 * sl_negotiate_uni():
 * Negotiate a SeedLink connection in uni-station mode and issue
 * the DATA command.  This is compatible with SeedLink Protocol
 * version 2 or greater.
 * If 'selectors' != 0 then the string is parsed on space and each
 * selector is sent.
 * If 'seqnum' != -1 then data is requested starting at seqnum.
 *
 * Returns -1 on errors, otherwise returns the link descriptor.
 ***************************************************************************/
int
sl_negotiate_uni (SLconnection * slconn)
{
  int sellen = 0;
  int bytesread = 0;
  int acceptsel = 0;		/* Count of accepted selectors */
  char *selptr;
  char sendstr[100];		/* A buffer for command strings */
  char readbuf[100];		/* A buffer for responses */
  SLstation *curstation;

  /* Point to the first station desc. */
  curstation = slconn->stations;

  selptr = curstation->selectors;

  /* Send the selector(s) and check the response(s) */
  if (curstation->selectors != 0)
    {
      while (1)
	{
	  selptr += sellen;
	  selptr += strspn (selptr, " ");
	  sellen = strcspn (selptr, " ");

	  if (sellen == 0)
	    break;		/* end of while loop */

	  else if (sellen > SELSIZE)
	    {
	      sl_log (0, 0, "[%s] invalid selector: %.*s\n", slconn->sladdr,
		      sellen, selptr);
	      selptr += sellen;
	    }
	  else
	    {

	      /* Build SELECT command, send it and receive response */
	      sprintf (sendstr, "SELECT %.*s\r", sellen, selptr);
	      sl_log (0, 2, "[%s] sending: SELECT %.*s\n", slconn->sladdr,
		      sellen, selptr);
	      bytesread = sl_writedata (slconn->link, (void *) sendstr,
					strlen (sendstr) + 1, slconn->sladdr,
					readbuf, sizeof (readbuf));
	      if (bytesread < 0)
		{		/* Error from sl_writedata() */
		  return -1;
		}

	      /* Check response to SELECT */
	      if (!strncmp (readbuf, "OK\r\n", bytesread) && bytesread == 4)
		{
		  sl_log (0, 1, "[%s] selector %.*s is OK\n", slconn->sladdr,
			  sellen, selptr);
		  acceptsel++;
		}
	      else if (!strncmp (readbuf, "ERROR\r\n", bytesread) &&
		       bytesread == 7)
		{
		  sl_log (1, 0, "[%s] selector %.*s not accepted\n",
			  slconn->sladdr, sellen, selptr);
		}
	      else
		{
		  sl_log (1, 0,
			  "[%s] invalid response to SELECT command: %.*s\n",
			  slconn->sladdr, bytesread, readbuf);
		  return -1;
		}
	    }
	}

      /* Fail if none of the given selectors were accepted */
      if (!acceptsel)
	{
	  sl_log (1, 0, "[%s] No data stream selector(s) accepted",
		  slconn->sladdr);
	  return -1;
	}
      else
	{
	  sl_log (0, 1, "[%s] %d selector(s) accepted\n",
		  slconn->sladdr, acceptsel);
	}
    }				/* End of selector processing */

  /* Issue the DATA or TIME action commands.  A specified start (and optionally,
     stop time) takes precedence over the resumption from any previous sequence
     number. */
  if (slconn->begin_time != NULL)
    {
      if (sl_checkversion (slconn, 2.92) >= 0)
	{
	  if (slconn->end_time == NULL)
	    {
	      sprintf (sendstr, "TIME %.25s\r", slconn->begin_time);
	    }
	  else
	    {
	      sprintf (sendstr, "TIME %.25s %.25s\r", slconn->begin_time,
		       slconn->end_time);
	    }
	  sl_log (0, 1, "[%s] requesting specified time window\n",
		  slconn->sladdr);
	}
      else
	{
	  sl_log (1, 0,
		  "[%s] detected SeedLink version (%.3f) does not support TIME windows\n",
		  slconn->sladdr, slconn->server_version);
	}
    }
  else if (curstation->seqnum != -1)
    {
      /* Increment sequence number by 1 */
      sprintf (sendstr, "DATA %06X\r", (curstation->seqnum + 1) & 0xffffff);
      sl_log (0, 1, "[%s] resuming data from %06X (Dec %d)\n",
	      slconn->sladdr, (curstation->seqnum + 1) & 0xffffff,
	      (curstation->seqnum + 1));
    }
  else
    {
      sprintf (sendstr, "DATA\r");
      sl_log (0, 1, "[%s] requesting next available data\n", slconn->sladdr);
    }

  if (sl_writedata (slconn->link, (void *) sendstr, strlen (sendstr) + 1,
		    slconn->sladdr, (void *) NULL, 0) < 0)
    {
      sl_log (1, 0, "[%s] error sending DATA/TIME request\n", slconn->sladdr);
      return -1;
    }

  return slconn->link;
}				/* End of sl_configlink() */


/***************************************************************************
 * sl_negotiate_multi():
 * Negotiate a SeedLink connection using multi-station mode and 
 * issue the END action command.  This is compatible with SeedLink
 * Protocol version 3, multi-station mode.
 * If 'curstation->selectors' != 0 then the string is parsed on space
 * and each selector is sent.
 * If 'curstation->seqnum' != -1 then data is requested starting at
 * seqnum.
 *
 * Returns -1 on errors, otherwise returns the link descriptor.
 ***************************************************************************/
int
sl_negotiate_multi (SLconnection * slconn)
{
  int sellen = 0;
  int bytesread = 0;
  int acceptsta = 0;		/* Count of accepted stations */
  int acceptsel = 0;		/* Count of accepted selectors */
  char *selptr;
  char sendstr[100];		/* A buffer for command strings */
  char readbuf[100];		/* A buffer for responses */
  char slring[12];		/* Keep track of the ring name */
  SLstation *curstation;

  /* Point to the first station desc. */
  curstation = slconn->stations;

  /* Loop through the station chain */
  while (curstation != NULL)
    {

      /* A ring identifier */
      snprintf (slring, sizeof (slring), "%s_%s",
		curstation->net, curstation->sta);

      /* Send the STATION command */
      sprintf (sendstr, "STATION %s %s\r", curstation->sta, curstation->net);
      sl_log (0, 2, "[%s] sending: STATION %s %s\n",
	      slring, curstation->sta, curstation->net);
      bytesread = sl_writedata (slconn->link, (void *) sendstr,
				strlen (sendstr) + 1, slring, readbuf,
				sizeof (readbuf));
      if (bytesread < 0)
	{
	  return -1;
	}

      /* Check the response */
      if (!strncmp (readbuf, "OK\r\n", bytesread) && bytesread == 4)
	{
	  sl_log (0, 2, "[%s] station is OK (selected)\n", slring);
	  acceptsta++;
	}
      else if (!strncmp (readbuf, "ERROR\r\n", bytesread) && bytesread == 7)
	{
	  sl_log (1, 0, "[%s] station not accepted, skipping\n", slring);
	  /* Increment the loop control and skip to the next station */
	  curstation = curstation->next;
	  continue;
	}
      else
	{
	  sl_log (1, 0, "[%s] invalid response to STATION command: %.*s\n",
		  slring, bytesread, readbuf);
	  return -1;
	}

      selptr = curstation->selectors;
      sellen = 0;

      /* Send the selector(s) and check the response(s) */
      if (curstation->selectors != 0)
	{
	  while (1)
	    {
	      selptr += sellen;
	      selptr += strspn (selptr, " ");
	      sellen = strcspn (selptr, " ");

	      if (sellen == 0)
		break;		/* end of while loop */

	      else if (sellen > SELSIZE)
		{
		  sl_log (0, 0, "[%s] invalid selector: %.*s\n",
			  slring, sellen, selptr);
		  selptr += sellen;
		}
	      else
		{

		  /* Build SELECT command, send it and receive response */
		  sprintf (sendstr, "SELECT %.*s\r", sellen, selptr);
		  sl_log (0, 2, "[%s] sending: SELECT %.*s\n", slring, sellen,
			  selptr);
		  bytesread =
		    sl_writedata (slconn->link, (void *) sendstr,
				  strlen (sendstr) + 1, slring, readbuf,
				  sizeof (readbuf));
		  if (bytesread < 0)
		    {		/* Error from sl_writedata() */
		      return -1;
		    }

		  /* Check response to SELECT */
		  if (!strncmp (readbuf, "OK\r\n", bytesread)
		      && bytesread == 4)
		    {
		      sl_log (0, 2, "[%s] selector %.*s is OK\n", slring,
			      sellen, selptr);
		      acceptsel++;
		    }
		  else if (!strncmp (readbuf, "ERROR\r\n", bytesread) &&
			   bytesread == 7)
		    {
		      sl_log (1, 0, "[%s] selector %.*s not accepted\n",
			      slring, sellen, selptr);
		    }
		  else
		    {
		      sl_log (1, 0,
			      "[%s] invalid response to SELECT command: %.*s\n",
			      slring, bytesread, readbuf);
		      return -1;
		    }
		}
	    }

	  /* Fail if none of the given selectors were accepted */
	  if (!acceptsel)
	    {
	      sl_log (1, 0, "[%s] No data stream selector(s) accepted",
		      slring);
	      return -1;
	    }
	  else
	    {
	      sl_log (0, 2, "[%s] %d selector(s) accepted\n", slring,
		      acceptsel);
	    }

	  acceptsel = 0;	/* Reset the accepted selector count */

	}			/* End of selector processing */

      /* Issue the DATA or TIME action commands.  A specified start (and optionally,
         stop time) takes precedence over the resumption from any previous sequence
         number. */
      if (slconn->begin_time != NULL)
	{
	  if (sl_checkversion (slconn, 2.92) >= 0)
	    {
	      if (slconn->end_time == NULL)
		{
		  sprintf (sendstr, "TIME %.25s\r", slconn->begin_time);
		}
	      else
		{
		  sprintf (sendstr, "TIME %.25s %.25s\r", slconn->begin_time,
			   slconn->end_time);
		}
	      sl_log (0, 1, "[%s] requesting specified time window\n",
		      slring);
	    }
	  else
	    {
	      sl_log (1, 0,
		      "[%s] detected SeedLink version (%.3f) does not support TIME windows\n",
		      slring, slconn->server_version);
	    }
	}
      else if (curstation->seqnum != -1)
	{
	  /* Increment sequence number by 1 */
	  sprintf (sendstr, "DATA %06X\r",
		   (curstation->seqnum + 1) & 0xffffff);
	  sl_log (0, 1, "[%s] resuming data from %06X (Dec %d)\n", slring,
		  (curstation->seqnum + 1) & 0xffffff,
		  (curstation->seqnum + 1));
	}
      else
	{
	  sprintf (sendstr, "DATA\r");
	  sl_log (0, 1, "[%s] requesting next available data\n", slring);
	}

      /* Send the DATA/TIME command and receive response */
      bytesread = sl_writedata (slconn->link, (void *) sendstr,
				strlen (sendstr) + 1, slring, readbuf,
				sizeof (readbuf));
      if (bytesread < 0)
	{
	  sl_log (1, 0, "[%s] error with DATA/TIME request\n", slring);
	  return -1;
	}

      /* Check response to DATA/TIME request */
      if (!strncmp (readbuf, "OK\r\n", bytesread) && bytesread == 4)
	{
	  sl_log (0, 2, "[%s] DATA/TIME command is OK\n", slring);
	}
      else if (!strncmp (readbuf, "ERROR\r\n", bytesread) && bytesread == 7)
	{
	  sl_log (1, 0, "[%s] DATA/TIME command is not accepted\n", slring);
	}
      else
	{
	  sl_log (1, 0, "[%s] invalid response to DATA/TIME command: %.*s\n",
		  slring, bytesread, readbuf);
	  return -1;
	}

      /* Point to the next station */
      curstation = curstation->next;

    }				/* End of station and selector config (end of station chain). */

  /* Fail if no stations were accepted */
  if (!acceptsta)
    {
      sl_log (1, 0, "[%s] No station(s) accepted\n", slconn->sladdr);
      return -1;
    }
  else
    {
      sl_log (0, 1, "[%s] %d station(s) accepted\n",
	      slconn->sladdr, acceptsta);
    }

  /* Issue END action command */
  sprintf (sendstr, "END\r");
  sl_log (0, 2, "[%s] sending: END\n", slconn->sladdr);
  if (sl_writedata (slconn->link, (void *) sendstr, strlen (sendstr) + 1,
		    slconn->sladdr, (void *) NULL, 0) < 0)
    {
      sl_log (1, 0, "[%s] error sending END command\n", slconn->sladdr);
      return -1;
    }

  return slconn->link;
}				/* End of sl_configlink_multi() */


/***************************************************************************
 * sl_request_info():
 * Send a request for the specified INFO level.  The verbosity level can
 * be specified, allowing control of when the request should be logged.
 *
 * Returns -1 on errors, otherwise the socket descriptor created.
 ***************************************************************************/
int
sl_request_info (SLconnection * slconn, const char * info_level,
		 int verb_level)
{
  char sendstr[40];		/* A buffer for command strings */

  if (sl_checkversion (slconn, 2.92) >= 0)
    {
      sprintf (sendstr, "INFO %.15s\r", info_level);

      sl_log (0, verb_level, "[%s] requesting INFO level %s\n",
	      slconn->sladdr, info_level);

      if (sl_writedata (slconn->link, (void *) sendstr, strlen (sendstr) + 1,
			slconn->sladdr, (void *) NULL, 0) < 0)
	{
	  sl_log (1, 0, "[%s] error sending INFO request\n", slconn->sladdr);
	  return -1;
	}
    }
  else
    {
      sl_log (1, 0,
	      "[%s] detected SeedLink version (%.3f) does not support INFO requests\n",
	      slconn->sladdr, slconn->server_version);
      return -1;
    }

  return slconn->link;
}				/* End of sl_request_info() */


/***************************************************************************
 * sl_connect():
 * Open a network socket connection to a SeedLink server and set
 * 'slconn->link' to the new descriptor.  Expects 'slconn->sladdr'
 * to be in 'host:port' format.
 *
 * Returns -1 on errors, otherwise the socket descriptor created.
 ***************************************************************************/
int
sl_connect (SLconnection * slconn)
{
  int sock;
  int on = 1;
  int fdflags;
  int sockstat;
  unsigned short int host_port;
  char *host_name, *ptr, *tail;
  struct hostent *hostinfo;
  struct sockaddr_in inet_addr;

  if ((host_name = strdup (slconn->sladdr)) == NULL)
    {
      sl_log (1, 0, "strdup(): cannot copy SeedLink address string\n");
      return -1;
    }				/* ? needed? a hang over from the original */

  if ((ptr = strchr (host_name, ':')) == NULL)
    {
      sl_log (1, 0, "[%s] host address is not in `[hostname]:port' format\n",
	      slconn->sladdr);
      return -1;
    }
  *ptr++ = '\0';

  host_port = strtoul (ptr, &tail, 0);
  if (*tail || host_port > 0xffff)
    {
      sl_log (1, 0, "[%s] host address is not in `[hostname]:port' format\n",
	      slconn->sladdr);
      return -1;
    }

  if ((hostinfo = gethostbyname (host_name)) == NULL)
    {
      /* unknown host, maybe temporary nameserver outage */
      sl_log (0, 0, "[%s] cannot resolve hostname: %s\n",
	      slconn->sladdr, host_name);
      return -1;
    }

  free (host_name);

  memset (&inet_addr, 0, sizeof (inet_addr));
  inet_addr.sin_family = AF_INET;
  inet_addr.sin_port = htons (host_port);
  inet_addr.sin_addr = *(struct in_addr *) hostinfo->h_addr_list[0];

  if ((sock = socket (PF_INET, SOCK_STREAM, 0)) < 0)
    {
      sl_log (1, 0, "[%s] socket(): %s\n", slconn->sladdr, strerror (errno));
      close (sock);
      return -1;
    }

  /* Set non-blocking IO */
  if ((fdflags = fcntl (sock, F_GETFL)) >= 0)
    fcntl (sock, F_SETFL, fdflags | O_NONBLOCK);

  if ((connect (sock, (struct sockaddr *) &inet_addr, sizeof (inet_addr))) ==
      -1)
    {
      if (errno != EINPROGRESS)
	{
	  sl_log (1, 0, "[%s] connect(): %s\n", slconn->sladdr,
		  strerror (errno));
	  close (sock);
	  return -1;
	}
    }

  /* Give a second for connecting, this is stupid but needed at the moment.
     Some firewalls that proxy/NAT TCP connections will report that the TCP
     connection is ready even when they are still opening the "other" end.
     Maybe a new test should be implemented to test for readiness.
   */
  sleep (1);

  /* Wait up to 10 seconds for the socket to be connected */
  if ((sockstat = sl_checksock (sock, 10, 0)) <= 0)
    {
      if (sockstat < 0)
	{			/* select() returned error */
	  sl_log (1, 1, "[%s] socket connect error\n", slconn->sladdr);
	}
      else
	{			/* socket time-out */
	  sl_log (0, 1, "[%s] socket connect time-out (10s)\n",
		  slconn->sladdr);
	}
      close (sock);
      return -1;
    }
  else
    {				/* socket connected */
      sl_log (0, 1, "[%s] network socket opened\n", slconn->sladdr);

      /* Set the SO_KEEPALIVE socket option, not really useful in this case */
      if (setsockopt
	  (sock, SOL_SOCKET, SO_KEEPALIVE, (char *) &on, sizeof (on)) < 0)
	sl_log (0, 1, "[%s] cannot set SO_KEEPALIVE socket option\n",
		slconn->sladdr);

      slconn->link = sock;

      /* Everything should be connected, say hello */
      if (sl_sayhello (slconn) == -1)
	{
	  close (sock);
	  return -1;
	}

      return sock;
    }
}				/* End of sl_connect() */


/***************************************************************************
 * sl_disconnect():
 * Close the network socket associated with connection
 *
 * Returns -1 on errors, otherwise the socket descriptor created.
 ***************************************************************************/
int
sl_disconnect (SLconnection * slconn)
{
  /* Shutdown() is not really necessary (?), but no harm done */

  /* Only shutdown() if connected */
  if (sl_checksock (slconn->link, 0, 0) > 1)
    {
      if (shutdown (slconn->link, 2) < 0)
	sl_log (1, 1, "[%s] shutdown(%d) failed: %s\n", slconn->sladdr,
		slconn->link, strerror (errno));
    }

  if (close (slconn->link) < 0)
    {
      sl_log (1, 1, "[%s] close(%d) failed: %s\n", slconn->sladdr,
	      slconn->link, strerror (errno));
    }

  sl_log (0, 1, "[%s] network socket closed\n", slconn->sladdr);

  return -1;			/* can be used to set the old descriptor */
}				/* End of sl_disconnect() */


/***************************************************************************
 * sl_checksock():
 * Check a socket for write ability using select() and read ability
 * using recv(... MSG_PEEK).  Time-out values are also passed (seconds
 * and microseconds) for the select() call.
 *
 * Returns:
 *  1 = success
 *  0 = if time-out expires
 * -1 = errors
 ***************************************************************************/
int
sl_checksock (int sock, int tosec, int tousec)
{
  int sret;
  int ret = -1;			/* default is failure */
  char testbuf[1];
  fd_set checkset;
  struct timeval to;

  FD_ZERO (&checkset);
  FD_SET (sock, &checkset);

  to.tv_sec = tosec;
  to.tv_usec = tousec;

  /* Check write ability with select() */
  if ((sret = select (sock + 1, NULL, &checkset, NULL, &to)) > 0)
    ret = 1;
  else if (sret == 0)
    ret = 0;			/* time-out expired */

  /* Check read ability with recv() */
  if (ret && (recv (sock, testbuf, sizeof (char), MSG_PEEK)) <= 0)
    {
      if (errno == EAGAIN)
	ret = 1;		/* no data for non-blocking IO */
      else
	ret = -1;
    }

  return ret;
}				/* End of sl_checksock() */


/***************************************************************************
 * sl_sequence():
 * Check for 'SL' signature and sequence number.
 *
 * Returns the packet sequence number of the SeedLink packet.
 ***************************************************************************/
int
sl_sequence (const Seedlink_head * datapack)
{
  int seqnum;
  char seqstr[7], *sqtail;

  if (strncmp (datapack->seq, SIGNATURE, 2))
    return -1;
  strncpy (seqstr, &datapack->seq[2], 6);
  seqstr[6] = '\0';
  seqnum = strtoul (seqstr, &sqtail, 16);
  if (seqnum & ~0xffffff || *sqtail)
    return -1;
  return seqnum;
}				/* End of sl_sequence() */


/***************************************************************************
 * sl_blocktype():
 * Check the type of first 'important' blockette found in data record if
 * any at all.
 *
 * Returns the block type.
 ***************************************************************************/
int
sl_blocktype (const Seedlink_head * datapack)
{
  int b2000 = 0;

  int16_t   samples_in_record;
  int16_t   sample_rate_factor;
  int16_t   first_blockette_byte;
  uint16_t blockette_type;
  uint16_t next_offset;

  const Blockette_head *p;

  samples_in_record    = (int16_t) ntohs(datapack->seedhead.samples_in_record);
  sample_rate_factor   = (int16_t) ntohs(datapack->seedhead.sample_rate_factor);
  first_blockette_byte = (int16_t) ntohs(datapack->seedhead.first_blockette_byte);

  p = (const Blockette_head *) ((const char *) &datapack->seedhead + 
				first_blockette_byte);

  blockette_type       = (uint16_t) ntohs(p->blockette_type);
  next_offset          = (uint16_t) ntohs(p->next_offset);

  do
    {
      if (((const char *) p) - ((const char *) &datapack->seedhead) >
	  MAX_HEADER_LEN)
	{
	  return BTNUM;
	}

      if (blockette_type >= 200 && blockette_type <= 299)
	return BTDET;
      if (blockette_type >= 300 && blockette_type <= 399)
	return BTCAL;
      if (blockette_type >= 500 && blockette_type <= 599)
	return BTTIM;
      if (blockette_type == 2000)
	b2000 = 1;

      p = (const Blockette_head *) ((const char *) &datapack->seedhead +
				    next_offset);
      
      blockette_type       = (uint16_t) ntohs(p->blockette_type);
      next_offset          = (uint16_t) ntohs(p->next_offset);
    }
  while ((const Seed_record_header *) p != &datapack->seedhead);

  if (sample_rate_factor == 0)
    {
      if (samples_in_record != 0)
	return BTMSG;
      if (b2000)
	return BTBLK;
    }

  return BTDATA;
}				/* End of sl_blocktype() */


/***************************************************************************
 * sl_dtime();
 * Return a double precision version of the current time since the epoch.
 * This is the number of seconds since Jan. 1, 1970 without leap seconds.
 ***************************************************************************/
double
sl_dtime (void)
{
  struct timeval tp;
  struct timezone tzp;

  gettimeofday (&tp, &tzp);
  return ((double) tp.tv_sec + ((double) tp.tv_usec / 1000000.0));
}				/* End of sl_dtime() */


/***************************************************************************
 * sl_writedata():
 * write() 'buflen' bytes from 'buffer' to 'writelink'.
 * 'code' is a string to include in error messages for identification.
 * If 'resp' is not NULL then read up to 'resplen' bytes into 'resp'
 * after sending 'buffer'.  This is only designed for small pieces of data,
 * specifically the server responses to commands.
 *
 * Returns -1 on error, and size (in bytes) of the response
 * received (0 if 'resp' == NULL).
 ***************************************************************************/
int
sl_writedata (int writelink, void *buffer, int buflen, char *code,
	      void *resp, int resplen)
{

  int bytesread = 0;		/* bytes read in the read loop */
  int ackcnt = 0;		/* counter for the read loop */
  int ackpoll = 50000;		/* poll at 0.05 seconds for reading */

  if (write (writelink, buffer, buflen) < 0)
    {
      sl_log (1, 0, "[%s] error sending '%.*s '\n", code,
	      strcspn ((char *) buffer, "\r\n"), (char *) buffer);
      return -1;
    }

  /* If requested, wait up to 30 seconds for a response */
  if (resp != NULL)
    {
      while ((bytesread = sl_readdata (writelink, resp, resplen, code)) == 0)
	{
	  if (ackcnt > 600)
	    {			/* 30 second wait, (ackpoll x 600) */
	      sl_log (1, 0,
		      "[%s] no response from SeedLink server for '%.*s'\n",
		      code, strcspn ((char *) buffer, "\r\n"),
		      (char *) buffer);
	      return -1;
	    }
	  usleep (ackpoll);
	  ackcnt++;
	}
    }

  if (bytesread == -1)
    {
      sl_log (1, 0, "[%s] bad response to '%.20s'\n", code, (char *) buffer);
    }

  return bytesread;
}				/* End of sl_writedata() */


/***************************************************************************
 * sl_readdata():
 * read() 'maxbytes' data from 'readlink' into a specified 'buffer'.
 * 'code' is a string to be included in error messages for identification.
 *
 * Returns -1 on error/EOF, 0 for no available data and the number
 * of bytes read on success.
 ***************************************************************************/
int
sl_readdata (int readlink, void *buffer, size_t maxbytes, char *code)
{
  int bytesread = 0;

  bytesread = read (readlink, buffer, maxbytes);

  if (bytesread == 0)		/* should indicate TCP FIN or EOF */
    {
      sl_log (1, 1, "[%s] read():%d TCP FIN or EOF received\n", code,
	      bytesread);
      return -1;
    }
  else if (bytesread < 0)
    {
      if (errno != EAGAIN)
	{
	  sl_log (1, 0, "[%s] read():%d %s\n", code, bytesread,
		  strerror (errno));
	  return -1;
	}
      /* no data available for NONBLOCKing IO */
      return 0;
    }

  return bytesread;
}				/* End of sl_readdata() */


/***************************************************************************
 * sl_log():
 * A standard logging/printing routine
 * This function works in two modes:
 * 1 - Initialization, expecting 2 arguments with the first (level)
 *     being -1 and the second being verbosity.  This will set the
 *     verbosity for all future calls, the default is 0.  Can be used
 *     to change the verbosity at any time. I.e. 'sl_log(-1,2);'
 *
 * 2 - expecting 3+ arguments, log level, verbosity level, printf
 *     format, and printf arguments.  If the verbosity level is less
 *     than or equal to the set verbosity (see mode 1), the printf
 *     format and arguments will be printed at the appropriate log
 *     level. I.e. 'sl_log(0, 0, "error: %s", result);'
 *
 * Returns the new verbosity if using mode 1.
 * Returns the number of characters formatted on success, and a
 *   a negative value on error if using mode 2.
 *
 * This routine acts as a central message facility for the other SeedLink
 * functions and can be modified to use another message handler if
 * needed.
 ***************************************************************************/
int
sl_log (int level, int verb, ...)
{
  static int staticverb = 0;
  int retvalue;

  if (level == -1)
    {
      staticverb = verb;
      retvalue = staticverb;
    }
  else if (verb <= staticverb)
    {
      char message[100];
      char *format;
      va_list listptr;

      va_start (listptr, verb);
      format = va_arg (listptr, char *);
      retvalue = vsnprintf (message, 100, format, listptr);

      if (level == 1)
	{
	  /*
	  printf("error: %s",message);
	  */
	  elog_complain(0,"%s", message);
	}
      else
	{
	  /*
	  printf("%s", message);
	  */
	  elog_notify(0,"%s", message);
	}

      va_end (listptr);
    }

  return retvalue;
}				/* End of sl_log() */
