
#ifndef SLINKUTILS_H
#define SLINKUTILS_H 1

#include <stdio.h>
#include <stdlib.h>
#include <stdarg.h>
#include <unistd.h>
#include <errno.h>
#include <fcntl.h>
#include <string.h>
#include <unistd.h>
#include <netdb.h>
#include <sys/types.h>
#include <sys/wait.h>
#include <sys/time.h>
#include <sys/socket.h>
#include <netinet/in.h>

#define PACKSIZE        512
#define MAX_HEADER_LEN  64
#define SEQSIZE         8
#define SELSIZE         8
#define SIGNATURE       "SL"
#define BUFSIZE         4096

/* Blockette types */
#define BTDATA 0		/* waveform data, miniSEED record */
#define BTDET 1			/* detection blockette */
#define BTCAL 2			/* calibration blockette */
#define BTTIM 3			/* timing blockette */
#define BTMSG 4			/* message blockette */
#define BTBLK 5			/* general blockette */
#define BTNUM 6			/* used as the error indicator (same as BTCHAN) */
#define BTCHA 6			/* for requesting channel info or detectors */

#define BTXML 10		/* an XML formatted message in a miniSEED log
				   record, used for STATUS and keepalive/heartbeat 
				   responses */

/* Common fields within SEED headers and other structures */
typedef char Seed_name_type[3];
typedef char Seed_net_type[2];
typedef char Location_type[2];

/* SEED time structure (10 bytes) */
typedef struct seed_time_s
{
  int16_t yr;	     	        /* year */
  int16_t jday;		        /* julian day */
  unsigned char hr;		/* hour */
  unsigned char minute;		/* minutes */
  unsigned char seconds;	/* seconds */
  unsigned char unused;		/* zero */
  int16_t tenth_millisec;	/* tenth of a millisecond */
}
Seed_time;

/* SEED data record header (48 bytes) */
typedef struct seed_record_header_s
{
  char sequence[6];		        /* record number */
  char seed_record_type;
  char continuation_record;
  char station_ID_call_letters[5];
  Location_type location_id;	        /* non aligned! */
  Seed_name_type channel_id;	        /* non aligned! */
  Seed_net_type seednet;
  Seed_time starting_time;
  int16_t samples_in_record;
  int16_t sample_rate_factor;
  int16_t sample_rate_multiplier;
  unsigned char activity_flags;	        /* ?I?LEBTC */
  unsigned char IO_flags;	        /* ??L????? */
  unsigned char data_quality_flags;	/* Q??M???? */
  unsigned char number_of_following_blockettes;
  int32_t tenth_msec_correction;	/* temporarily 0 */
  int16_t first_data_byte;	        /* 0, 48, or 56 or multiple of 64 */
  int16_t first_blockette_byte;	        /* 0 or 48 */
}
Seed_record_header;

/* SEED blockette header (4 bytes) */
typedef struct blockette_head_s
{
  uint16_t blockette_type;
  uint16_t next_offset;
}
Blockette_head;

/* SeedLink header (56 bytes) */
typedef struct seedlink_head_s
{
  char seq[SEQSIZE];
  Seed_record_header seedhead;
}
Seedlink_head;

/* Station information */
typedef struct slstation_s
{
  char   *net;		        /* The network code */
  char   *sta;		        /* The station code */
  char   *selectors;	        /* SeedLink style selectors for this station */
  int    seqnum;	        /* SeedLink sequence number for this station */
  struct slstation_s *next;     /* The next station in the chain */
}
SLstation;

/* Server connection information */
typedef struct slconnection_s
{
  char       *sladdr;           /* The host:port of a SeedLink server */
  int        link;		/* The link descriptor */
  float      server_version;    /* Version of the remote SeedLink server */
  short int  multistation;      /* Boolean flag to indicate multistation */
  const char *begin_time;	/* Used with TIME request */
  const char *end_time;		/* Used with TIME request */
  SLstation  *stations;		/* Pointer to a station desc. struct */
}
SLconnection;


/*
 * Declarations for functions in slinkutils.c
 */

extern double sl_dtime (void);
extern int sl_configlink (SLconnection * slconn);
extern int sl_request_info (SLconnection * slconn, const char * info_level,
			    int verb_level);
extern int sl_connect (SLconnection * slconn);
extern int sl_disconnect (SLconnection * slconn);
extern int sl_checksock (int sock, int tosec, int tousec);
extern int sl_sequence (const Seedlink_head *);
extern int sl_blocktype (const Seedlink_head *);
extern int sl_writedata (int writelink, void *buffer, int buflen,
			 char *code, void *resp, int resplen);
extern int sl_readdata (int readlink, void *buffer, size_t maxbytes,
			char *code);
extern int sl_log (int level, int verb, ...);

#endif /* SLINKUTILS_H */
